
(** Distributed structures *)

module type TYPE =
 sig
  type 'a t

  val init : int -> (int -> 'a) -> 'a t
  val length : 'a t -> int
  val empty : 'a t
  val map : ('a -> 'b) -> 'a t -> 'b t
  val zip : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
  val zip2 : ('a -> 'b -> 'c -> 'd) -> 'a t -> 'b t -> 'c t -> 'd t
  val to_array : 'a t -> 'a array
  val to_list : 'a t -> 'a list
  val shift_left : 'a -> 'a t -> 'a t
  val shift_right : 'a -> 'a t -> 'a t
  val concat : 'a t list -> 'a t
  val get_last : 'a t -> 'a
  val get_first : 'a t -> 'a
  val sub : 'a t -> int -> int -> 'a t
 end

module Make : functor (T : TYPE) -> functor (Bsml : Bsmlsig.BSML) ->
 sig

  (** [make f n] gives a parallel vector holding a distributed
      structure of size n, with f applied onto each index of the
      structure (written [[...]]):  

      < [[ f 0 ; f 1 ; ... ; f (n-1) ]] , 
        [[ f n ; f (n+1) ; ... ]] , ... >.  

      The parallel cost is {i max (w{_i})} where {i w{_i}} is the
      execution time of {i f i}. *)
   val make : (int -> 'a) -> int -> 'a T.t Bsml.par

  (** [length] gives the size of the distributed structure.
      The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the structure at process i. *)
  val length : 'a T.t Bsml.par -> int

 (** [extract] gathers the data from the distributed structure into an array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the structure at process i. *)
  val extract : 'a T.t Bsml.par -> 'a T.t

 (** [par] is the dual of extract, it distributes a sequential structure over many processors. *)
  val par : 'a T.t -> 'a T.t Bsml.par

 (** [map] applies a function of arity 1 on each element of the distributed structure.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map : ('a -> 'b) -> 'a T.t Bsml.par -> 'b T.t Bsml.par

 (** [zip] applies a function of arity 2 on each pair (element of the first
     distributed structure, element of the second distributed structure).
     The parallel cost is {i 3 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip : ('a -> 'b -> 'c) -> 'a T.t Bsml.par -> 'b T.t Bsml.par -> 'c T.t Bsml.par

 (** [map_index] applies a function of arity 2 (first argument is the index within
     the distributed structure) on each element of the distributed structure.
     The parallel cost is {i 5 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map_index : (int -> 'a -> 'b) -> 'a T.t Bsml.par -> 'b T.t Bsml.par

 (** [zip_index] applies a function of arity 3 (first argument is the index within
     the distributed structure on each pair (element of the first distributed structure,
     element of the second distributed structure).
     The parallel cost is {i 6 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip_index : (int -> 'a -> 'b -> 'c) -> 'a T.t Bsml.par -> 'b T.t Bsml.par -> 'c T.t Bsml.par

 (** [to_array] converts a distributed structure into a distributed array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the structure at process i. *)
  val to_array : 'a T.t Bsml.par -> 'a array Bsml.par

 (** [to_list] converts a distributed structure into a distributed list.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the structure at process i. *)
  val to_list : 'a T.t Bsml.par -> 'a list Bsml.par

 (** [shift_left v a] shifts the distributed structure a to the left and inserts v to its end.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the structure
     at process i and h = 1. *)
  val shift_left : 'a -> 'a T.t Bsml.par -> 'a T.t Bsml.par

 (** [shift_right v a] shifts the distributed structure to the right and inserts v to its beginning.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the structure
     at process i and h = 1. *)
  val shift_right : 'a -> 'a T.t Bsml.par -> 'a T.t Bsml.par
 end

module MakeArray : functor (Bsml : Bsmlsig.BSML) ->
 sig

 (** [make f n] gives a parallel vector holding a distributed array of size n,
     with f applied onto each index of the array.
      < \[|f 0 ; f 1 ; ... ; f (n-1)|\] , \[|f n ; f (n+1) ; ...|\] , ... >.
     The parallel cost is {i max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val make : (int -> 'a) -> int -> 'a array Bsml.par

 (** [length] gives the size of the distributed array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the array at process i. *)
  val length : 'a array Bsml.par -> int

 (** [extract] gathers the data from the distributed array into an array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the array at process i. *)
  val extract : 'a array Bsml.par -> 'a array

 (** [par] is the dual of extract, it distributes a sequential array over many processors. *)
  val par : 'a array -> 'a array Bsml.par
 
 (** [map] applies a function of arity 1 on each element of the distributed array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map : ('a -> 'b) -> 'a array Bsml.par -> 'b array Bsml.par

 (** [zip] applies a function of arity 2 on each pair (element of the first
     distributed array, element of the second distributed array).
     The parallel cost is {i 3 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip : ('a -> 'b -> 'c) -> 'a array Bsml.par -> 'b array Bsml.par -> 'c array Bsml.par

 (** [map_index] applies a function of arity 2 (first argument is the index within
     the distributed array) on each element of the distributed array.
     The parallel cost is {i 5 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map_index : (int -> 'a -> 'b) -> 'a array Bsml.par -> 'b array Bsml.par

 (** [zip_index] applies a function of arity 3 (first argument is the index within
     the distributed array on each pair (element of the first distributed array,
     element of the second distributed array).
     The parallel cost is {i 6 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip_index : (int -> 'a -> 'b -> 'c) -> 'a array Bsml.par -> 'b array Bsml.par -> 'c array Bsml.par

 (** [to_list] converts a distributed array into a distributed list.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the array at process i. *)
  val to_list : 'a array Bsml.par -> 'a list Bsml.par

 (** [shift_left v a] shifts the distributed array a to the left and inserts v to its end.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the array
     at process i and h = 1. *)
  val shift_left : 'a -> 'a array Bsml.par -> 'a array Bsml.par 

 (** [shift_right v a] shifts the distributed array to the right and inserts v to its beginning.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the array
     at process i and h = 1. *)
  val shift_right : 'a -> 'a array Bsml.par -> 'a array Bsml.par
 end

module MakeList : functor (Bsml : Bsmlsig.BSML) ->
 sig

 (** [make f n] gives a parallel vector holding a distributed list of size n,
      with f applied onto each index of the list.
      < \[f 0 ; f 1 ; ... ; f (n-1)\] , \[f n ; f (n+1) ; ...\] , ... >.
     The parallel cost is {i max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val make : (int -> 'a) -> int -> 'a list Bsml.par

 (** [length] gives the size of the distributed list.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the list at process i. *)
  val length : 'a list Bsml.par -> int

 (** [extract] gathers the data from the distributed list into a list.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the list at process i. *)
  val extract : 'a list Bsml.par -> 'a list

 (** [par] is the dual of extract, it distributes a sequential list over many processors. *)
  val par : 'a list -> 'a list Bsml.par

 (** [map] applies a function of arity 1 on each element of the distributed list.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map : ('a -> 'b) -> 'a list Bsml.par -> 'b list Bsml.par

 (** [zip] applies a function of arity 2 on each pair (element of the first
     distributed list, element of the second distributed list).
     The parallel cost is {i 3 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip : ('a -> 'b -> 'c) -> 'a list Bsml.par -> 'b list Bsml.par -> 'c list Bsml.par

 (** [map_index] applies a function of arity 2 (first argument is the index within
     the distributed list) on each element of the distributed list.
     The parallel cost is {i 5 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val map_index : (int -> 'a -> 'b) -> 'a list Bsml.par -> 'b list Bsml.par

 (** [zip_index] applies a function of arity 3 (first argument is the index within
     the distributed list on each pair (element of the first distributed list,
     element of the second distributed list).
     The parallel cost is {i 6 max (w{_i})} where {i w{_i}} is the execution time of {i f i}. *)
  val zip_index : (int -> 'a -> 'b -> 'c) -> 'a list Bsml.par -> 'b list Bsml.par -> 'c list Bsml.par

 (** [to_array] converts a distributed list into a distributed array.
     The parallel cost is {i 2 max (w{_i})} where {i w{_i}} is the exploring time of the list at process i. *)
  val to_array : 'a list Bsml.par -> 'a array Bsml.par
 
 (** [shift_left v a] shifts the distributed list a to the left and inserts v to its end.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the list
     at process i and h = 1. *)
  val shift_left : 'a -> 'a list Bsml.par -> 'a list Bsml.par 

 (** [shift_right v a] shifts the distributed list to the right and inserts v to its beginning.
     The parallel cost is {i 16 max (w{_i}) + g h + L} where {i w{_i}} is the exploring time of the list
     at process i and h = 1. *)
  val shift_right : 'a -> 'a list Bsml.par -> 'a list Bsml.par
 end
